/* Center Markdown Tables (requires md_in_html extension) */
.center-table {
    text-align: center;
}

.md-typeset .center-table :is(td,th):not([align]) {
    /* Reset alignment for table cells */
    text-align: initial;
}
